@@ -382,8 +382,7 @@ pgfnames(const char *path)
 
 	filenames = (char **) palloc(fnsize * sizeof(char *));
 
-	errno = 0;
-	while ((file = readdir(dir)) != NULL)
+	while (errno = 0, (file = readdir(dir)) != NULL)
 	{
 		if (strcmp(file->d_name, ".") != 0 && strcmp(file->d_name, "..") != 0)
 		{
@@ -395,17 +394,14 @@ pgfnames(const char *path)
 			}
 			filenames[numnames++] = pstrdup(file->d_name);
 		}
-		errno = 0;
 	}
-#ifdef WIN32
 
-	/*
-	 * This fix is in mingw cvs (runtime/mingwex/dirent.c rev 1.4), but not in
-	 * released version
-	 */
+#ifdef WIN32
+	/* Bug in old Mingw dirent.c;  fixed in mingw-runtime-3.2, 2003-10-10 */
 	if (GetLastError() == ERROR_NO_MORE_FILES)
 		errno = 0;
 #endif
+
 	if (errno)
 	{
 #ifndef FRONTEND
@@ -418,7 +414,15 @@ pgfnames(const char *path)
 
 	filenames[numnames] = NULL;
 
-	closedir(dir);
+	if (closedir(dir))
+	{
+#ifndef FRONTEND
+		elog(WARNING, "could not close directory \"%s\": %m", path);
+#else
+		fprintf(stderr, _("could not close directory \"%s\": %s\n"),
+				path, strerror(errno));
+#endif
+	}
 
 	return filenames;
 }
