@@ -13,6 +13,7 @@ use Project;
 use Solution;
 use Cwd;
 use File::Copy;
+use File::Basename;
 use Config;
 use VSObjectFactory;
 use List::Util qw(first);
@@ -24,23 +25,31 @@ our (@ISA, @EXPORT_OK);
 
 my $solution;
 my $libpgport;
+my $libpgcommon;
 my $postgres;
 my $libpq;
 
 my $contrib_defines = { 'refint' => 'REFINT_VERBOSE' };
 my @contrib_uselibpq =
-  ('dblink', 'oid2name', 'pgbench', 'pg_upgrade', 'vacuumlo');
+  ('dblink', 'oid2name', 'pgbench', 'pg_upgrade', 'postgres_fdw', 'vacuumlo');
 my @contrib_uselibpgport = (
 	'oid2name',      'pgbench',
 	'pg_standby',    'pg_archivecleanup',
 	'pg_test_fsync', 'pg_test_timing',
-	'pg_upgrade',    'vacuumlo');
+	'pg_upgrade',    'pg_xlogdump',
+	'vacuumlo');
+my @contrib_uselibpgcommon = (
+	'oid2name',      'pgbench',
+	'pg_standby',    'pg_archivecleanup',
+	'pg_test_fsync', 'pg_test_timing',
+	'pg_upgrade',    'pg_xlogdump',
+	'vacuumlo');
 my $contrib_extralibs = { 'pgbench' => ['wsock32.lib'] };
 my $contrib_extraincludes =
   { 'tsearch2' => ['contrib/tsearch2'], 'dblink' => ['src/backend'] };
 my $contrib_extrasource = {
 	'cube' => [ 'cubescan.l', 'cubeparse.y' ],
-	'seg'  => [ 'segscan.l',  'segparse.y' ] };
+	'seg'  => [ 'segscan.l',  'segparse.y' ], };
 my @contrib_excludes = ('pgcrypto', 'intagg', 'sepgsql');
 
 sub mkvcbuild
@@ -59,14 +68,28 @@ sub mkvcbuild
 	  chklocale.c crypt.c fls.c fseeko.c getrusage.c inet_aton.c random.c
 	  srandom.c getaddrinfo.c gettimeofday.c inet_net_ntop.c kill.c open.c
 	  erand48.c snprintf.c strlcat.c strlcpy.c dirmod.c exec.c noblock.c path.c
-	  pgcheckdir.c pg_crc.c pgmkdirp.c pgsleep.c pgstrcasecmp.c qsort.c qsort_arg.c
-	  sprompt.c tar.c thread.c getopt.c getopt_long.c dirent.c rint.c win32env.c
-	  win32error.c win32setlocale.c mkdtemp.c);
+	  pgcheckdir.c pg_crc.c pgmkdirp.c pgsleep.c pgstrcasecmp.c pqsignal.c
+	  mkdtemp.c qsort.c qsort_arg.c quotes.c
+	  sprompt.c tar.c thread.c wait_error.c getopt.c getopt_long.c dirent.c
+	  win32env.c win32error.c win32setlocale.c);
+
+	push(@pgportfiles, 'rint.c') if ($vsVersion < '12.00');
+
+	our @pgcommonallfiles = qw(
+	  relpath.c);
+
+	our @pgcommonfrontendfiles = (@pgcommonallfiles, qw(fe_memutils.c));
+
+	our @pgcommonbkndfiles = @pgcommonallfiles;
 
 	$libpgport = $solution->AddProject('libpgport', 'lib', 'misc');
 	$libpgport->AddDefine('FRONTEND');
 	$libpgport->AddFiles('src\port', @pgportfiles);
 
+	$libpgcommon = $solution->AddProject('libpgcommon', 'lib', 'misc');
+	$libpgcommon->AddDefine('FRONTEND');
+	$libpgcommon->AddFiles('src\common', @pgcommonfrontendfiles);
+
 	$postgres = $solution->AddProject('postgres', 'exe', '', 'src\backend');
 	$postgres->AddIncludeDir('src\backend');
 	$postgres->AddDir('src\backend\port\win32');
@@ -80,7 +103,8 @@ sub mkvcbuild
 		'src\backend\port\win32_shmem.c');
 	$postgres->ReplaceFile('src\backend\port\pg_latch.c',
 		'src\backend\port\win32_latch.c');
-	$postgres->AddFiles('src\port', @pgportfiles);
+	$postgres->AddFiles('src\port',   @pgportfiles);
+	$postgres->AddFiles('src\common', @pgcommonbkndfiles);
 	$postgres->AddDir('src\timezone');
 	$postgres->AddFiles('src\backend\parser', 'scan.l', 'gram.y');
 	$postgres->AddFiles('src\backend\bootstrap', 'bootscanner.l',
@@ -107,7 +131,7 @@ sub mkvcbuild
 
 	my $plpgsql =
 	  $solution->AddProject('plpgsql', 'dll', 'PLs', 'src\pl\plpgsql\src');
-	$plpgsql->AddFiles('src\pl\plpgsql\src', 'gram.y');
+	$plpgsql->AddFiles('src\pl\plpgsql\src', 'pl_gram.y');
 	$plpgsql->AddReference($postgres);
 
 	if ($solution->{options}->{perl})
@@ -116,26 +140,7 @@ sub mkvcbuild
 		my $plperl =
 		  $solution->AddProject('plperl', 'dll', 'PLs', 'src\pl\plperl');
 		$plperl->AddIncludeDir($solution->{options}->{perl} . '/lib/CORE');
-
-		# Add defines from Perl's ccflags; see PGAC_CHECK_PERL_EMBED_CCFLAGS
-		my @perl_embed_ccflags;
-		foreach my $f (split(" ",$Config{ccflags}))
-		{
-			if ($f =~ /^-D[^_]/)
-			{
-				$f =~ s/\-D//;
-				push(@perl_embed_ccflags, $f);
-			}
-		}
-
-		# XXX this probably is redundant now?
-		push(@perl_embed_ccflags, 'PLPERL_HAVE_UID_GID');
-
-		foreach my $f (@perl_embed_ccflags)
-		{
-			$plperl->AddDefine($f);
-		}
-
+		$plperl->AddDefine('PLPERL_HAVE_UID_GID');
 		foreach my $xs ('SPI.xs', 'Util.xs')
 		{
 			(my $xsc = $xs) =~ s/\.xs/.c/;
@@ -244,24 +249,20 @@ sub mkvcbuild
 
 	if ($solution->{options}->{tcl})
 	{
-		my $found = 0;
 		my $pltcl =
 		  $solution->AddProject('pltcl', 'dll', 'PLs', 'src\pl\tcl');
 		$pltcl->AddIncludeDir($solution->{options}->{tcl} . '\include');
 		$pltcl->AddReference($postgres);
-
-		for my $tclver (qw(86t 86 85 84))
+		if (-e $solution->{options}->{tcl} . '\lib\tcl85.lib')
 		{
-			my $tcllib = $solution->{options}->{tcl} . "\\lib\\tcl$tclver.lib";
-			if (-e $tcllib)
-			{
-				$pltcl->AddLibrary($tcllib);
-				$found = 1;
-				last;
-			}
+			$pltcl->AddLibrary(
+				$solution->{options}->{tcl} . '\lib\tcl85.lib');
+		}
+		else
+		{
+			$pltcl->AddLibrary(
+				$solution->{options}->{tcl} . '\lib\tcl84.lib');
 		}
-		die "Unable to find $solution->{options}->{tcl}\\lib\\tcl<version>.lib"
-			unless $found;
 	}
 
 	$libpq = $solution->AddProject('libpq', 'dll', 'interfaces',
@@ -305,7 +306,6 @@ sub mkvcbuild
 	my $libecpgcompat = $solution->AddProject(
 		'libecpg_compat', 'dll',
 		'interfaces',     'src\interfaces\ecpg\compatlib');
-	$libecpgcompat->AddDefine('FRONTEND');
 	$libecpgcompat->AddIncludeDir('src\interfaces\ecpg\include');
 	$libecpgcompat->AddIncludeDir('src\interfaces\libpq');
 	$libecpgcompat->UseDef('src\interfaces\ecpg\compatlib\compatlib.def');
@@ -318,10 +318,10 @@ sub mkvcbuild
 	$ecpg->AddPrefixInclude('src\interfaces\ecpg\preproc');
 	$ecpg->AddFiles('src\interfaces\ecpg\preproc', 'pgc.l', 'preproc.y');
 	$ecpg->AddDefine('MAJOR_VERSION=4');
-	$ecpg->AddDefine('MINOR_VERSION=2');
-	$ecpg->AddDefine('PATCHLEVEL=1');
+	$ecpg->AddDefine('MINOR_VERSION=9');
+	$ecpg->AddDefine('PATCHLEVEL=0');
 	$ecpg->AddDefine('ECPG_COMPILE');
-	$ecpg->AddReference($libpgport);
+	$ecpg->AddReference($libpgport, $libpgcommon);
 
 	my $pgregress_ecpg =
 	  $solution->AddProject('pg_regress_ecpg', 'exe', 'misc');
@@ -331,8 +331,7 @@ sub mkvcbuild
 	$pgregress_ecpg->AddIncludeDir('src\test\regress');
 	$pgregress_ecpg->AddDefine('HOST_TUPLE="i686-pc-win32vc"');
 	$pgregress_ecpg->AddDefine('FRONTEND');
-	$pgregress_ecpg->AddLibrary('ws2_32.lib');
-	$pgregress_ecpg->AddReference($libpgport);
+	$pgregress_ecpg->AddReference($libpgport, $libpgcommon);
 
 	my $isolation_tester =
 	  $solution->AddProject('isolationtester', 'exe', 'misc');
@@ -357,8 +356,7 @@ sub mkvcbuild
 	$pgregress_isolation->AddIncludeDir('src\test\regress');
 	$pgregress_isolation->AddDefine('HOST_TUPLE="i686-pc-win32vc"');
 	$pgregress_isolation->AddDefine('FRONTEND');
-	$pgregress_isolation->AddLibrary('ws2_32.lib');
-	$pgregress_isolation->AddReference($libpgport);
+	$pgregress_isolation->AddReference($libpgport, $libpgcommon);
 
 	# src/bin
 	my $initdb = AddSimpleFrontend('initdb');
@@ -397,6 +395,7 @@ sub mkvcbuild
 	$psql->AddIncludeDir('src\bin\pg_dump');
 	$psql->AddIncludeDir('src\backend');
 	$psql->AddFile('src\bin\psql\psqlscan.l');
+	$psql->AddLibrary('ws2_32.lib');
 
 	my $pgdump = AddSimpleFrontend('pg_dump', 1);
 	$pgdump->AddIncludeDir('src\backend');
@@ -405,6 +404,7 @@ sub mkvcbuild
 	$pgdump->AddFile('src\bin\pg_dump\pg_dump_sort.c');
 	$pgdump->AddFile('src\bin\pg_dump\keywords.c');
 	$pgdump->AddFile('src\backend\parser\kwlookup.c');
+	$pgdump->AddLibrary('ws2_32.lib');
 
 	my $pgdumpall = AddSimpleFrontend('pg_dump', 1);
 
@@ -419,9 +419,9 @@ sub mkvcbuild
 	$pgdumpall->AddIncludeDir('src\backend');
 	$pgdumpall->AddFile('src\bin\pg_dump\pg_dumpall.c');
 	$pgdumpall->AddFile('src\bin\pg_dump\dumputils.c');
-	$pgdumpall->AddFile('src\bin\pg_dump\dumpmem.c');
 	$pgdumpall->AddFile('src\bin\pg_dump\keywords.c');
 	$pgdumpall->AddFile('src\backend\parser\kwlookup.c');
+	$pgdumpall->AddLibrary('ws2_32.lib');
 
 	my $pgrestore = AddSimpleFrontend('pg_dump', 1);
 	$pgrestore->{name} = 'pg_restore';
@@ -429,10 +429,12 @@ sub mkvcbuild
 	$pgrestore->AddFile('src\bin\pg_dump\pg_restore.c');
 	$pgrestore->AddFile('src\bin\pg_dump\keywords.c');
 	$pgrestore->AddFile('src\backend\parser\kwlookup.c');
+	$pgrestore->AddLibrary('ws2_32.lib');
 
 	my $zic = $solution->AddProject('zic', 'exe', 'utils');
-	$zic->AddFiles('src\timezone', 'zic.c');
-	$zic->AddReference($libpgport);
+	$zic->AddFiles('src\timezone', 'zic.c', 'ialloc.c', 'scheck.c',
+		'localtime.c');
+	$zic->AddReference($libpgport, $libpgcommon);
 
 	if ($solution->{options}->{xml})
 	{
@@ -523,8 +525,8 @@ sub mkvcbuild
 		my $mf = Project::read_file(
 			'src\backend\utils\mb\conversion_procs\\' . $sub . '\Makefile');
 		my $p = $solution->AddProject($sub, 'dll', 'conversion procs');
-		$p->AddFile('src\backend\utils\mb\conversion_procs\\' 
-			  . $sub . '\\' 
+		$p->AddFile('src\backend\utils\mb\conversion_procs\\'
+			  . $sub . '\\'
 			  . $sub
 			  . '.c');
 		if ($mf =~ m{^SRCS\s*\+=\s*(.*)$}m)
@@ -572,8 +574,9 @@ sub mkvcbuild
 		$proj->AddIncludeDir('src\interfaces\libpq');
 		$proj->AddIncludeDir('src\bin\pg_dump');
 		$proj->AddIncludeDir('src\bin\psql');
-		$proj->AddReference($libpq, $libpgport);
+		$proj->AddReference($libpq, $libpgport, $libpgcommon);
 		$proj->AddResourceFile('src\bin\scripts', 'PostgreSQL Utility');
+		$proj->AddLibrary('ws2_32.lib');
 	}
 
 	# Regression DLL and EXE
@@ -586,9 +589,23 @@ sub mkvcbuild
 	$pgregress->AddFile('src\test\regress\pg_regress_main.c');
 	$pgregress->AddIncludeDir('src\port');
 	$pgregress->AddDefine('HOST_TUPLE="i686-pc-win32vc"');
-	$pgregress->AddDefine('FRONTEND');
-	$pgregress->AddLibrary('ws2_32.lib');
-	$pgregress->AddReference($libpgport);
+	$pgregress->AddReference($libpgport, $libpgcommon);
+
+	# fix up pg_xlogdump once it's been set up
+	# files symlinked on Unix are copied on windows
+	my $pg_xlogdump =
+	  (grep { $_->{name} eq 'pg_xlogdump' }
+		  @{ $solution->{projects}->{contrib} })[0];
+	$pg_xlogdump->AddDefine('FRONTEND');
+	foreach my $xf (glob('src/backend/access/rmgrdesc/*desc.c'))
+	{
+		my $bf = basename $xf;
+		copy($xf, "contrib/pg_xlogdump/$bf");
+		$pg_xlogdump->AddFile("contrib\\pg_xlogdump\\$bf");
+	}
+	copy(
+		'src/backend/access/transam/xlogreader.c',
+		'contrib/pg_xlogdump/xlogreader.c');
 
 	$solution->Save();
 	return $solution->{vcver};
@@ -606,7 +623,7 @@ sub AddSimpleFrontend
 
 	my $p = $solution->AddProject($n, 'exe', 'bin');
 	$p->AddDir('src\bin\\' . $n);
-	$p->AddReference($libpgport);
+	$p->AddReference($libpgport, $libpgcommon);
 	if ($uselibpq)
 	{
 		$p->AddIncludeDir('src\interfaces\libpq');
@@ -757,6 +774,10 @@ sub AdjustContribProj
 	{
 		$proj->AddReference($libpgport);
 	}
+	if (grep { /^$n$/ } @contrib_uselibpgcommon)
+	{
+		$proj->AddReference($libpgcommon);
+	}
 	if ($contrib_extralibs->{$n})
 	{
 		foreach my $l (@{ $contrib_extralibs->{$n} })
